/*****************************************************
Program for measuring inductance, capacitance and frequency. Frequency is
measured using timer 0 with an external clock source, being the frequency
to be measured. This increments TCNT0, and after an overflow, counter.
Timer 1 is used to generate a 1 second time base, after which the value in
counter + TCNT give a direct frequency readout. The inductance and secondary
capacitance measuring functions use the difference in frequency between a
reference value found during program start-up and the new resonant frequency
caused by the reactive component under test.

For measuring above 100nF the effect of resistance on test the capacitor's charging time
is used. The time it takes to reach one time constant (63.2% of full voltage) is tau = RC.
So C = tau / R. R is known, and tau is measured with timer 0. With a regualted 5V supply,
one time constant is equivilant to:

     0.632 * 5.00 = 3,16V 
     
The analog comparator on PortB is used to trigger an interrupt when the test
capacitor has charged to one time constant. The reference voltage of 3,16V
should be applied to pin 4, and the test capacitor to pin 3.

Program has two capacitance ranges, which are automatically selected. Range 1 uses
39 k resistor and is used between 50nF and 1100nF. Range 2 uses 470 Ohm resistor and is
used between 1F and up. The maximun capacitance is measureable is 90mF, after which the
counter variable will overflow. Keep in mind the time constant is 42s at 90mF, meaning
it would take near a minute to measure...
I'll leave it up to you to optimize for quicker read out if desireable.

The program starts by draining a test capacitor, then charging it through a known
resistance. During the charge process Timer 0 increments a variable. Once the capacitor
has charged for what equals a time constant the compartor triggers an interrupt. The
interrupt rutine does the calculation of the capacitance. The the capacitance is processed
before being displayed on the LCD.

Data Types:

int8_t - Signed Char 
uint16_t - Unsigned Int 
uint32_t - Unsigned Long 
int64_t - Signed Long Long

(signed/unsigned) char - 1 byte 
(signed/unsigned) short - 2 bytes 
(signed/unsigned) int - 2 bytes 
(signed/unsigned) long - 4 bytes 
(signed/unsigned) long long - 8 bytes 
float - 4 bytes (floating point) 
double - alias to float


Project : LCF Meter
AVR Clock frequency: 8,00 MHz
*****************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/delay.h>
#include "lcd.h"

#define c_ref 1.000								// Depends on reference capacitor
#define twoPI 6.2831853
#define scaling 1E15

#define r_high 470.0       						// size of high range resistor in Ohms
#define r_low 39.0								// size of low range resistor in kOhms
#define tau_c 31.875							// Timer 0 interrupt period in s


/*	Ignore these, just used for reference
#define mode1 PORTB.PIN1
#define mode2 PORTB.PIN4
#define drain PORTB.5         					// active high
#define low_range PORTB.6     					// active low
#define high_range PORTB.7    					// active low
*/

// Declare global variables
volatile uint32_t counter;             			// increments with every timer 0 overflow
volatile uint32_t frequency;           			// current frequency measurment
volatile double reactance;						// used to calculate reactances in all calculations
		 uint8_t test;							// indicates which mode the meter is in
volatile uint8_t range;							// if set meter is in high range mode (uses small resistor)
volatile uint8_t testing;						// used to lock program in testing loop


// Function for displaying long integers, only used frequency meter
void display(uint32_t output)
{
    char line_txt[16];
    unsigned int display_lsd;
    
    if (output < 1000){
        sprintf(line_txt, "Fx = %lu Hz",output);
    }else if (output < 1000000){
        display_lsd = output % 1000;
        output = output / 1000;
        sprintf(line_txt, "Fx = %lu.%03u kHz",output, display_lsd);
    }else{
        display_lsd = (output / 1000) % 1000;
        output = output / 1000000;
        sprintf(line_txt, "Fx = %lu.%03u MHz",output, display_lsd);
    };  
    
    lcd_clrscr();
    lcd_puts(line_txt);
}

// Function for calculating capacitance, using frequency deviance
void measure_c(void)
{
	char line_txt[16];
	char *suffix;
	static uint32_t f_ref;						// Reference Frequency

// check that capacitance is in valid range
if (frequency > 18000)		// about 890nF, you should change over to RC by this point anyway
	{
	
	// set fixed capacitance
	if (f_ref == 0)
	{
		f_ref = frequency;
	};
	
	// math out capacitance
	reactance = ((double)(f_ref) / frequency);
	reactance = c_ref * ((reactance * reactance) - 1);
	
	// formatt and display capacitance
	if (reactance > 1.0)
		{
		suffix = "nF";
	}else{
		reactance = reactance * 1000;
		suffix = "pF";
	}
	
	sprintf(line_txt, "Cx = %.2f %s",reactance,suffix);
    lcd_clrscr();
    lcd_puts(line_txt);
	
// if not, error
}else
	{
    lcd_clrscr();
    lcd_puts("Over Range!");
};

	//display((uint32_t)(reactance*1000));
}

// Function for calculating inductance
void measure_l(void)
{
	char line_txt[16];
	char *suffix;
	static double l_ref;							// Reference Inductance

// check that inductance is in valid range
if (frequency > 1000)
	{
	// if so, math out inductance
	reactance = (twoPI*frequency);
	reactance = c_ref * (reactance * reactance);
	reactance =  scaling / reactance;
	reactance = reactance - l_ref;
	
	// set value of fixed inductance
	if (l_ref == 0){
		l_ref = reactance;
	}
	
	// formatt and display inductance
	if (reactance > 1000)
		{
		reactance = reactance / 1000;
		suffix = "mH";
	}else if (reactance < 1.0){
		reactance = reactance * 1000;
		suffix = "nH";
	}else{
		suffix = "uH";
	}
	
	sprintf(line_txt, "Lx = %.2f %s",reactance,suffix);
    lcd_clrscr();
    lcd_puts(line_txt);
// if not, error
}else
	{
    lcd_clrscr();
    lcd_puts("Over Range!");
};
	
	//display((uint32_t)(reactance*1000));
}

// Timer 0 overflow interrupt service routine
ISR(TIMER0_OVF_vect)
{
    counter++;

}

// Timer1 overflow interrupt service routine
ISR(TIMER1_OVF_vect)
{
// Disable Interrupts
	asm volatile("cli");
    
	// frequency is total number of Timer 0 counts.
    frequency = (255 * counter) + TCNT0;

// start up delay for three cycles
	if (range < 3)
	{
		range++;
	};

// check that start-up phase is over
if (range > 2){
	// check metering mode and go to corresponding function
	if (test == 16)
		{
		measure_c();
	}else if (test == 0)
		{
		measure_l();
	}else if (test == 2)
		{
		display(frequency);
	};
};
    

// Set Timer 1 start point for 1 second interrupt rate
    TCNT1H=0xE17B >> 8;
    TCNT1L=0xE17B & 0xff;
// Reinitialize Timer values, Enable Interrupts
    counter = 0;
    TCNT0 = 0;
	asm volatile("sei");

}

// Interrupt Rutine for analoge comparator
ISR(ANA_COMP_vect)
{
    // Grab current timer 0 value, minus delay (which is constant)
    reactance = TCNT0 - 37;
	
	// disable interrupts and turn drain capacitor
	asm volatile("cli");
	PORTB = 0b11100000;
    
    // release program from testing loop
    testing = 0;
    
}

// Charge test capacitor and measure time, calculate capacitance
void measure(void)
{
	char output;
	char line_txt[16];
	char *suffix;

    // Pick measuring range
    if (range){
		output = 0b01000000;
    }else{
		output = 0b10000000;
    };
	
    testing = 1;
	
	// Global enable interrupts 
    asm volatile("sei");
	
    // turn off capacitor grounding
    PORTB = output;
	counter = 0;
	TCNT0 = 0;
	
    // wait for analog comparator to trip
    while (testing);
	
	// do teh math

	// get precentage of Timer 0 register overflow when stopped
	reactance = (reactance / 255.0) + counter;
	
	// filter out spurious math errors
	reactance = reactance * 1000;
	reactance = (uint32_t)(reactance / 1000);
	
	// Format and calculate capacitance
	if (range){
		reactance = (reactance * tau_c) / r_high;
		suffix = "uF";
	}else{
		reactance = (reactance * tau_c) / r_low;
		suffix = "nF";
	};
	
	// set range for next time
	if (reactance == 0){
		range = 1;
	}else if (counter < 9000){
		range = 0;
	};
	
	// display hard earned capacitance
	sprintf(line_txt,"Cx = %.2f %s",reactance,suffix);
	lcd_clrscr();
	lcd_puts(line_txt);
	_delay_ms(500);
	
}

// Main
void main(void)
{

// Port B initialization
// Func7=Out Func6=Out Func5=Out Func4=In Func3=In Func2=In Func1=In Func0=In 
// State7=0 State6=0 State5=0 State4=P State3=T State2=T State1=P State0=T 
PORTB=0x12;
DDRB=0xE0;

// Port C initialization
// Func7=Out Func6=Out Func5=Out Func4=Out Func3=In Func2=Out Func1=Out Func0=Out 
// State7=0 State6=0 State5=0 State4=0 State3=T State2=0 State1=0 State0=0 
PORTC=0x00;
DDRC=0xF7;

// Get metering mode
test = (PINB & 0b000010010);

/* Testing mode gathered from inputs
	16 Measure capacitance using frequency deviance 	- PIN4 high, PIN1 low
	0 Mesure inductance using frequency deviance 		- PIN4 low, PIN1 low
	2 Measure frequency 								- PIN4 low, PIN1 high
	18 Measure capacitance using time constant 			- PIN4 high, PIN1 high
*/

// Declaration of Timers
if (test == 18)
{
// Declarations for RC meter
// Timer/Counter 0 initialization
// Clock source: System Clock
// Clock value: 8000,000 kHz
// Mode: Normal top=FFh
// OC0 output: Disconnected
TCCR0=0x01;
TCNT0=0x00;
OCR0=0x00;

// Timer(s)/Counter(s) Interrupt(s) initialization
TIMSK=0x02;

// Analog Comparator initialization
// Analog Comparator: On
// Interrupt on Rising Output Edge
// Analog Comparator Input Capture by Timer/Counter 1: Off
ACSR=0x0B;

}else{
// Declarations for frequency counter based meter
// Timer/Counter 0 initialization
// Clock source: T0 pin Rising Edge
TCCR0=0x07;
TCNT0=0x00;
OCR0=0x00;

// Timer/Counter 1 initialization
// Clock source: System Clock
// Clock value: 7,813 kHz
// Timer1 Overflow Interrupt: On
TCCR1A=0x00;
TCCR1B=0x05;
TCNT1H=0xE1;
TCNT1L=0x7B;
ICR1H=0x00;
ICR1L=0x00;
OCR1AH=0x00;
OCR1AL=0x00;
OCR1BH=0x00;
OCR1BL=0x00;

// Timer(s)/Counter(s) Interrupt(s) initialization
TIMSK=0x82;

// Analog Comparator initialization
// Analog Comparator: Off
// Analog Comparator Input Capture by Timer/Counter 1: Off
ACSR=0x80;
	
};

// Global disable interrupts
asm volatile("cli");

// LCD module initialization
lcd_init(LCD_DISP_ON);
lcd_clrscr();
lcd_puts("Initializing...");

// Run main loop for RC measuring meter
if (test == 18)
{
	//  drain test capacitor and turn off 5V supplies, set high range mode
	PORTB = 0b11100000;
	range = 1;
	_delay_ms(400);

	while (1)
    {
		measure();   
    };
};

// Global enable interrupts, wait for oscillator to stabilize
asm volatile("sei");

while (1)
	{
	
	};

}    
